"""Specific use of FuncNode and DAG"""

from dataclasses import fields
from inspect import signature
from typing import Callable, Union
from functools import partial

from i2 import Sig, kwargs_trans
from meshed.base import FuncNode, func_node_transformer
from meshed.dag import DAG
from meshed.util import Renamer, numbered_suffix_renamer, InvalidFunctionParameters

_func_node_fields = {x.name for x in fields(FuncNode)}

# TODO: How to enforce that it only has keys from _func_node_fields?
FuncNodeKwargs = dict
FuncNodeKwargsTrans = Callable[[FuncNodeKwargs], FuncNodeKwargs]


def is_func_node_kwargs_trans(func: Callable) -> bool:
    """Returns True iff the only required params of func are FuncNode field names.
    This ensures that the func will be able to be bound to FuncNode fields and
    therefore used as a func_node (kwargs) transformer.
    """
    return _func_node_fields.issuperset(Sig(func).required_names)


def func_node_kwargs_trans(func: Callable) -> FuncNodeKwargsTrans:
    if is_func_node_kwargs_trans(func):
        return func
    else:
        raise InvalidFunctionParameters(
            f'A FuncNodeKwargsTrans is expected to only have required params that are '
            f'also '
            f"FuncNode fields ({', '.join(_func_node_fields)}). "
            'Function {func} had signature: {Sig(func)}'
        )


def func_node_name_trans(
    name_trans: Callable[[str], Union[str, None]],
    *,
    also_apply_to_func_label: bool = False,
):
    """

    :param name_trans: A function taking a str and returning a str, or None (to indicate
        that no transformation should take place).
    :param also_apply_to_func_label:
    :return:
    """
    if not is_func_node_kwargs_trans(name_trans):
        if Sig(name_trans).n_required <= 1:

            def name_trans(name):
                return name_trans(name)

    kwargs_trans_kwargs = dict(name=name_trans)
    if also_apply_to_func_label:
        kwargs_trans_kwargs.update(func_label=name_trans)

    return partial(
        func_node_transformer,
        kwargs_transformers=partial(kwargs_trans, **kwargs_trans_kwargs),
    )


# TODO: Extract  ingress/egress boilerplate to wrapper
def suffix_ids(
    func_nodes,
    renamer: Union[Renamer, str] = numbered_suffix_renamer,
    *,
    also_apply_to_func_label: bool = False,
):
    if isinstance(func_nodes, DAG):
        egress = DAG
    else:
        egress = list
    if isinstance(renamer, str):
        suffix = renamer
        renamer = lambda name: f'{name}{suffix}'
    assert callable(suffix), f'suffix needs to be callable'
    func_node_trans = func_node_name_trans(
        renamer, also_apply_to_func_label=also_apply_to_func_label,
    )
    return egress(map(func_node_trans, func_nodes))


# ---------------------------------------------------------------------------------------
# lined, with meshed


def get_param(func):
    """
    Find the name of the parameter of a function with exactly one parameter.
    Raise an error if more or less parameters.
    :param func: callable, the function to inspect
    :return: str, the name of the single parameter of func
    """

    params = signature(func).parameters.keys()
    assert (
        len(params) == 1
    ), f"Your function has more than 1 parameter! Namely: {', '.join(params)}"
    for param in params:
        return param


def line_with_dag(*steps):
    """
    Emulate a Line object with a DAG
    :param steps: an iterable of callables, the steps of the pipeline. Each step should have exactly one parameter
    and the output of each step is fed into the next
    :return: a DAG instance computing the composition of all the functions in steps, in the provided order
    """

    step_counter = 0
    first_node = FuncNode(steps[0], out=f'step_{step_counter}')
    func_nodes = [first_node]
    for step in steps[1:]:
        step_node = FuncNode(
            step,
            out=f'step_{step_counter + 1}',
            bind={get_param(step): f'step_{step_counter}'},
        )
        step_counter += 1
        func_nodes.append(step_node)

    return DAG(func_nodes)



from meshed import DAG
from creek.automatas import BasicAutomata, mapping_to_transition_func
from typing import Callable, MutableMapping, Any, Mapping, Literal
from dataclasses import dataclass
from i2 import ch_names


Case = Any
Cases = Mapping[Case, Callable]


RecordingCommands = Literal['start', 'resume', 'stop']


def mk_test_objects():
    # from slang import fixed_step_chunker

    audio = range(100)
    audio_chk_size = 5
    # audio_chks = list(fixed_step_chunker(audio, chk_size=audio_chk_size))
    audio_chks = [
        audio[i : i + audio_chk_size] for i in range(0, len(audio), audio_chk_size)
    ]
    plc_values = [0, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0]

    return audio_chks, plc_values


@dataclass
class RecordingSwitchBoard:
    store: MutableMapping = None
    _current_key = None

    def start(self, key, chk):
        self._current_key = key
        self.store[key] = []
        self._append(chk)

    def resume(self, key, chk):
        print(f'resume called')
        self._append(chk)

    def stop(self, key, chk):
        self._append(chk)
        self._current_key = None

    def _append(self, chk):
        if self._current_key is None:
            raise ValueError('Cannot append without first starting recording.')
        self.store[self._current_key].extend(chk)

    @property
    def is_recording(self):
        return self._current_key is not None


@dataclass
class SimpleSwitchCase:
    """A functional implementation of thw switch-case control flow.
    Makes a callable that takes two arguments, a case and an input.

    >>> f = SimpleSwitchCase({'plus_one': lambda x: x + 1, 'times_two': lambda x: x * 2})
    >>> f('plus_one', 2)
    3
    >>> f('times_two', 2)
    4
    """

    cases: Mapping[Case, Callable]

    def __call__(self, case, input):
        func = self.cases.get(case, None)
        if func is None:
            raise ValueError(f'Case {case} not found.')
        return func(input)


def mk_simple_switch_case(
    cases: Cases, *, name: str = None, case_name: str = None, input_name: str = None
):
    """
    Makes a simple switch-case function, with optional naming control.
    """
    switch_case_func = SimpleSwitchCase(cases)
    switch_case_func = ch_names(
        switch_case_func, **dict(case=case_name, input=input_name)
    )
    if name is not None:
        switch_case_func.__name__ = name
    return switch_case_func


def mk_recorder_switch(
    store, *, mk_recorder: Callable[[MutableMapping], Any] = RecordingSwitchBoard
):
    recorder = mk_recorder(store)
    return mk_simple_switch_case(
        {
            'start': lambda key_and_chk: recorder.start(*key_and_chk),
            'resume': lambda key_and_chk: recorder.resume(*key_and_chk),
            'stop': lambda key_and_chk: recorder.stop(*key_and_chk),
            'waiting': lambda x: None,
        },
        name='recorder_switch',
        case_name='state',
        input_name='key_and_chk',
    )


def mk_transition_func(
    trans_func_mapping, initial_state,  # symbol_var_name: str,
):
    recording_state_transition_func = mapping_to_transition_func(
        trans_func_mapping, strict=False,
    )
    transitioner = BasicAutomata(
        transition_func=recording_state_transition_func, state=initial_state,
    )

    # @i2.ch_names(symbol=symbol_var_name)
    def transition(symbol):
        return transitioner.transition(symbol)

    # transition = transitioner.reset().transition

    return transition


# store = mk_recorder_switch(store)
trans_func_mapping = {
    ('waiting', 1): 'start',
    ('start', 0): 'resume',
    ('start', 1): 'stop',
    ('resume', 1): 'stop',
    ('stop', 0): 'waiting',
    ('stop', 1): 'start',
}

# debugging tools
logger = {
    'symbol': [],
    'state': [],
    'state_func': [],
    'transition_func': [],
    'recorder': [],
}

# TFunc = mk_transition_func(trans_func_mapping, "waiting")
dag = DAG.from_funcs(
    recorder_switch=lambda store: mk_recorder_switch(store),
    recorder_logger=lambda recorder_switch: logger['recorder'].append(
        id(recorder_switch)
    ),
    # debug = lambda recorder_switch: print(id(recorder_switch)),
    transition_func=lambda trans_func_mapping: mk_transition_func(
        trans_func_mapping, 'waiting'
    ),
    transition_logger=lambda transition_func: logger['transition_func'].append(
        transition_func
    ),
    symbol=lambda plc: plc,
    symbol_logger=lambda symbol: logger['symbol'].append(symbol),
    state=lambda transition_func, symbol: transition_func(symbol),
    # tFunc=lambda: TFunc,
    # state=lambda tFunc, symbol: tFunc(symbol),
    state_logger=lambda state: logger['state'].append(state),
    key_and_chk=lambda key, chk: (key, chk),
    # key_and_chk_logger=lambda key_and_chk: logger['key_and_chk'].append(key_and_chk),
    state_func=lambda recorder_switch, state, key_and_chk: recorder_switch(
        state, key_and_chk
    ),
    state_func_logger=lambda state_func: logger['state_func'].append(state_func),
    output=lambda state_func, key_and_chk: state_func(*key_and_chk)
    if state_func is not None
    else None,
    result=lambda recorder_switch, transition_func, symbol, state, key_and_chk, state_func, output: dict(
        recorder_switch=recorder_switch,
        transition_func=transition_func,
        symbol=symbol,
        state=state,
        key_and_chk=key_and_chk,
        state_func=state_func,
        output=output,
    ),
)


if __name__ == '__main__':
    store = dict()

    my_dag = dag.partial(store=store, trans_func_mapping=trans_func_mapping)
    # my_dag.dot_digraph()
    # print(i2.Sig(my_dag))

    audio_chks, plc_values = mk_test_objects()
    keys = range(max(len(audio_chks), len(plc_values)))  # need some source of keys now!

    for chk, plc, key in zip(audio_chks, plc_values, keys):
        # print(f"{store =}{chk=} {plc=} {key=} ")
        # print(f'{store=}')
        # print(f"{my_dag.last_scope=}")

        # print(f"{my_dag[:'state_func'](chk=chk, plc=plc, key=key)}")
        # res = my_dag(chk=chk, plc=plc, key=key)
        res = dag(
            store=store,
            trans_func_mapping=trans_func_mapping,
            chk=chk,
            plc=plc,
            key=key,
        )

        # print(store)  # Careful: use keyword
    print(logger['symbol'])

    print(logger['state'])
    print(logger['recorder'])

    # transitioner = logger['transition_func'][-1]
    transitioner = TFunc
    state_sequence = list(map(transitioner, plc_values))
    print(state_sequence)
    # print(transitioner)



"""Visualization utilities for the meshed package."""

from typing import Iterable, Any
from i2.signatures import Sig


def dot_lines_of_objs(objs: Iterable, start_lines=(), end_lines=(), **kwargs):
    r"""
    Get lines generator for the graphviz.DiGraph(body=list(...))

    >>> from meshed.base import FuncNode
    >>> def add(a, b=1):
    ...     return a + b
    >>> def mult(x, y=3):
    ...     return x * y
    >>> def exp(mult, a):
    ...     return mult ** a
    >>> func_nodes = [
    ...     FuncNode(add, out='x'),
    ...     FuncNode(mult, name='the_product'),
    ...     FuncNode(exp)
    ... ]
    >>> lines = list(dot_lines_of_objs(func_nodes))
    >>> assert lines == [
    ... 'x [label="x" shape="none"]',
    ... '_add [label="_add" shape="box"]',
    ... '_add -> x',
    ... 'a [label="a" shape="none"]',
    ... 'b [label="b=" shape="none"]',
    ... 'a -> _add',
    ... 'b -> _add',
    ... 'mult [label="mult" shape="none"]',
    ... 'the_product [label="the_product" shape="box"]',
    ... 'the_product -> mult',
    ... 'x [label="x" shape="none"]',
    ... 'y [label="y=" shape="none"]',
    ... 'x -> the_product',
    ... 'y -> the_product',
    ... 'exp [label="exp" shape="none"]',
    ... '_exp [label="_exp" shape="box"]',
    ... '_exp -> exp',
    ... 'mult [label="mult" shape="none"]',
    ... 'a [label="a" shape="none"]',
    ... 'mult -> _exp',
    ... 'a -> _exp'
    ... ]  # doctest: +SKIP

    >>> from meshed.util import dot_to_ascii
    >>>
    >>> print(dot_to_ascii('\n'.join(lines)))  # doctest: +SKIP
    <BLANKLINE>
                    a        ─┐
                              │
               │              │
               │              │
               ▼              │
             ┌─────────────┐  │
     b=  ──▶ │    _add     │  │
             └─────────────┘  │
               │              │
               │              │
               ▼              │
                              │
                    x         │
                              │
               │              │
               │              │
               ▼              │
             ┌─────────────┐  │
     y=  ──▶ │ the_product │  │
             └─────────────┘  │
               │              │
               │              │
               ▼              │
                              │
                  mult        │
                              │
               │              │
               │              │
               ▼              │
             ┌─────────────┐  │
             │    _exp     │ ◀┘
             └─────────────┘
               │
               │
               ▼
    <BLANKLINE>
                   exp
    <BLANKLINE>

    """
    # Should we validate here, or outside this module?
    # from meshed.base import validate_that_func_node_names_are_sane
    # validate_that_func_node_names_are_sane(func_nodes)
    yield from start_lines
    for obj in objs:
        yield from obj.dot_lines(**kwargs)
    yield from end_lines


dot_lines_of_func_nodes = dot_lines_of_objs  # backwards compatiblity alias


# TODO: Should we integrate this to dot_lines_of_func_parameters directly (decorator?)
def add_new_line_if_none(s: str):
    """Since graphviz 0.18, need to have a newline in body lines.
    This util is there to address that, adding newlines to body lines
    when missing."""
    if s and s[-1] != '\n':
        return s + '\n'
    return s


# ------------------------------------------------------------------------------
# Unused -- consider deleting
def _parameters_and_names_from_sig(
    sig: Sig, out=None, func_name=None,
):
    func_name = func_name or sig.name
    out = out or sig.name
    if func_name == out:
        func_name = '_' + func_name
    assert isinstance(func_name, str) and isinstance(out, str)
    return sig.parameters, out, func_name


# ------------------------------------------------------------------------------
# Old stuff


def visualize_graph(graph):
    import graphviz
    from IPython.display import display

    dot = graphviz.Digraph()

    # Add nodes to the graph
    for node in graph:
        dot.node(node)

    # Add edges to the graph
    for node, neighbors in graph.items():
        for neighbor in neighbors:
            dot.edge(node, neighbor)

    # Render and display the graph in the notebook
    display(dot)


def visualize_graph_interactive(graph):
    import graphviz

    import networkx as nx
    import ipywidgets as widgets
    from IPython.display import display

    g = nx.DiGraph(graph)

    # Create an empty Graphviz graph
    dot = graphviz.Digraph()

    # Add nodes to the Graphviz graph
    for node in g.nodes:
        dot.node(str(node))

    # Add edges to the Graphviz graph
    for edge in g.edges:
        dot.edge(str(edge[0]), str(edge[1]))

    # Render the initial graph visualization
    graph_widget = widgets.HTML(value=dot.pipe(format='svg').decode('utf-8'))
    display(graph_widget)

    def add_edge(sender):
        source = source_node.value
        target = target_node.value
        if (source, target) not in g.edges:
            g.add_edge(source, target)
            dot.edge(str(source), str(target))
            graph_widget.value = dot.pipe(format='svg').decode('utf-8')
        source_node.value = ''
        target_node.value = ''

    def add_node(sender):
        node = new_node.value
        if node not in g.nodes:
            g.add_node(node)
            dot.node(str(node))
            graph_widget.value = dot.pipe(format='svg').decode('utf-8')
        new_node.value = ''

    def delete_edge(sender):
        source = str(delete_source.value)
        target = str(delete_target.value)
        if (source, target) in g.edges:
            g.remove_edge(source, target)
            dot.body.remove(f'\t{source} -> {target}\n')
            graph_widget.value = dot.pipe(format='svg').decode('utf-8')
        delete_source.value = ''
        delete_target.value = ''

    def delete_node(sender):
        node = delete_node_value.value
        if node in g.nodes:
            g.remove_node(node)
            dot.body[:] = [line for line in dot.body if str(node) not in line]
            graph_widget.value = dot.pipe(format='svg').decode('utf-8')
        delete_node_value.value = ''

    source_node = widgets.Text(placeholder='Source Node')
    target_node = widgets.Text(placeholder='Target Node')
    add_edge_button = widgets.Button(description='Add Edge')
    add_edge_button.on_click(add_edge)

    new_node = widgets.Text(placeholder='New Node')
    add_node_button = widgets.Button(description='Add Node')
    add_node_button.on_click(add_node)

    delete_source = widgets.Text(placeholder='Source Node')
    delete_target = widgets.Text(placeholder='Target Node')
    delete_edge_button = widgets.Button(description='Delete Edge')
    delete_edge_button.on_click(delete_edge)

    delete_node_value = widgets.Text(placeholder='Node')
    delete_node_button = widgets.Button(description='Delete Node')
    delete_node_button.on_click(delete_node)

    controls = widgets.HBox([source_node, target_node, add_edge_button])
    controls2 = widgets.HBox([new_node, add_node_button])
    controls3 = widgets.HBox([delete_source, delete_target, delete_edge_button])
    controls4 = widgets.HBox([delete_node_value, delete_node_button])
    display(controls)
    display(controls2)
    display(controls3)
    display(controls4)



import pytest
from pytest import fixture
from meshed import DAG
from i2 import Sig


@fixture
def foo():
    def func(x, y=1):
        return x + y

    return func


def test_addition_variadics():
    def foo(w, /, x: float, y='YY', *, z: str = 'ZZ', **rest):
        pass

    sig = Sig(foo)
    # res = sig.map_arguments(
    #    (11, 22, "you"), dict(z="zoo", other="stuff"), post_process=True
    # )
    # assert res == "{'w': 11, 'x': 22, 'y': 'you', 'z': 'zoo', 'other': 'stuff'}"
    assert True



"""Tools to work with meshed"""

from collections import namedtuple
from contextlib import contextmanager
from functools import cached_property, partial
import multiprocessing
import os
import time
from typing import Callable, List
from urllib.parse import urljoin

import i2

from meshed.dag import DAG


HOST = os.environ.get('HOST', '0.0.0.0')
PORT = int(os.environ.get('PORT', 3030))
API_URL = os.environ.get('API_URL', f'http://localhost:{PORT}')
SERVER = os.environ.get('SERVER', 'wsgiref')
OPENAPI_URL = urljoin(API_URL, 'openapi')


def find_funcs(dag, func_outs):
    return list(dag.find_funcs(lambda x: x.out in func_outs))


def mk_dag_with_ws_funcs(dag: DAG, ws_funcs: dict) -> DAG:
    """Creates a new DAG with the web service functions.

    :param dag: DAG to be hybridized
    :type dag: DAG
    :param ws_funcs: mapping of web service functions
    :type ws_funcs: dict
    :return: new DAG with the web service functions
    :rtype: DAG
    """
    return dag.ch_funcs(**ws_funcs)


def launch_funcs_webservice(funcs: List[Callable]):
    """Launches a web service application with the specified functions.

    :param funcs: functions to be hosted by the web service
    :type funcs: List[Callable]
    """
    from extrude import mk_api, run_api

    ws_app = mk_api(funcs, openapi=dict(base_url=API_URL))
    run_api(ws_app, host=HOST, port=PORT, server=SERVER)


@contextmanager
def launch_webservice(funcs_to_cloudify, wait_after_start_seconds=10):
    """Context manager to launch a web service application in a separate process."""
    ws = multiprocessing.Process(
        target=launch_funcs_webservice, args=(funcs_to_cloudify,)
    )
    ws.start()
    # TODO: I prefer using a timeout instead of a fixed wait time
    # TODO: Use strand tool for this: https://github.com/i2mint/strand/blob/7443631e9d2486358f0a34ed182e85b6ded5e50c/strand/taskrunning/utils.py#L54
    time.sleep(wait_after_start_seconds)
    yield ws

    ws.terminate()


class CloudFunctions:
    def __init__(self, funcs: List[Callable], openapi_url=OPENAPI_URL, logger=print):
        """Creates a Python dictionary-like object that maps the web service functions to Python functions.

        :param funcs: list of functions hosted by the web service
        :type funcs: List[Callable]
        :param openapi_url: url to get openapi spec, defaults to "http://localhost:{PORT}/openapi"
        :type openapi_url: str, optional
        :param logger: logger function, defaults to print
        :type logger: Callable, optional
        """
        self.funcs = funcs
        self.openapi_url = openapi_url
        self.logger = logger if callable(logger) else lambda x: None

    @cached_property
    def http_client(self):
        from http2py import HttpClient

        try:
            return HttpClient(url=self.openapi_url)
        except Exception:
            self.logger(
                f'Could not connect to {self.openapi_url}. Waiting 10 seconds and trying again.'
            )
            time.sleep(10)
            return HttpClient(url=self.openapi_url)

    @cached_property
    def func_names(self):
        return frozenset(f.__name__ for f in self.funcs)

    def __getitem__(self, key):
        """Returns a Python function that calls the web service function.
        HttpClient is queried at the execution of the function.
        """

        @i2.Sig(next(f for f in self.funcs if key == f.__name__))
        def ws_func(*a, **kw):
            self.logger(f'Getting web service for: {key}')
            if (_wsf := getattr(self.http_client, key, None)) is not None:
                self.logger(f'Found web service for: {key}')
                return _wsf(*a, **kw)
            raise KeyError(key)

        return ws_func

    def __contains__(self, key):
        return key in self.func_names

    def __len__(self):
        return len(self.func_names)

    def keys(self):
        return self.func_names

    def values(self):
        return (self[k] for k in self.keys())

    def items(self):
        return ((k, self[k]) for k in self.keys())


def mk_hybrid_dag(dag: DAG, func_ids_to_cloudify: list):
    """Creates a hybrid DAG that uses the web service for the specified functions.

    :param dag: dag to be hybridized
    :type dag: DAG
    :param func_ids_to_cloudify: list of function ids to be cloudified
    :type func_ids_to_cloudify: list
    :return: namedtuple with funcs_to_cloudify, ws_dag and ws_funcs
    :rtype: namedtuple
    """
    funcs_to_cloudify = find_funcs(dag, func_ids_to_cloudify)
    ws_funcs = CloudFunctions(funcs_to_cloudify)
    ws_dag = mk_dag_with_ws_funcs(dag, ws_funcs)

    HybridDAG = namedtuple('HybridDAG', ['funcs_to_cloudify', 'ws_dag', 'ws_funcs'])
    return HybridDAG(funcs_to_cloudify, ws_dag, ws_funcs)
